/* FreeS/WAN config file parser (confread.c)
 * Copyright (C) 2001-2002 Mathieu Lafon - Arkoon Network Security
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * RCSID $Id: confread.c,v 1.8 2002/09/03 18:06:03 mathieu Exp $
 */

#include <stdlib.h>
#include <string.h>

#include "parser.h"
#include "confread.h"
#include "interfaces.h"
#include "log.h"

static char _tmp_err[512];

static void default_values (struct starter_config *cfg)
{
	if (!cfg) return;
	memset(cfg, 0, sizeof(struct starter_config));

	cfg->setup.fragicmp = TRUE;
	cfg->setup.hidetos = TRUE;
	cfg->setup.uniqueids = FALSE;

	cfg->conn_default.policy = POLICY_ENCRYPT | POLICY_TUNNEL | POLICY_PSK;

	cfg->conn_default.sa_ike_life_seconds = OAKLEY_ISAKMP_SA_LIFETIME_DEFAULT;
	cfg->conn_default.sa_ipsec_life_seconds = SA_LIFE_DURATION_DEFAULT;
	cfg->conn_default.sa_rekey_margin = SA_REPLACEMENT_MARGIN_DEFAULT;
	cfg->conn_default.sa_rekey_fuzz = SA_REPLACEMENT_FUZZ_DEFAULT;
	cfg->conn_default.sa_keying_tries = SA_REPLACEMENT_RETRIES_DEFAULT;

	cfg->conn_default.left.addr_family = AF_INET;
	anyaddr(AF_INET, &cfg->conn_default.left.addr);
	anyaddr(AF_INET, &cfg->conn_default.left.nexthop);
	cfg->conn_default.right.addr_family = AF_INET;
	anyaddr(AF_INET, &cfg->conn_default.right.addr);
	anyaddr(AF_INET, &cfg->conn_default.right.nexthop);

	cfg->conn_default.startup = STARTUP_NO;
	cfg->conn_default.state = STATE_LOADED;
}

#define ERR_FOUND(args...) \
	{ if (perr && (*perr==NULL)) { \
		snprintf(_tmp_err, sizeof(_tmp_err)-1, ## args); \
		*perr = strdup(_tmp_err); } \
	err++; }

#define KW_BOOL(str,val) else if (strcmp(kw->name,str)==0) { \
		if (strcmp(kw->value,"yes")==0) { val=TRUE; } \
		else if (strcmp(kw->value,"no")==0) { val=FALSE; } \
		else { ERR_FOUND("bad boolean value %s=%s", kw->name, kw->value) } \
	}

#define KW_POLICY_FLAG(str,sy,sn,fl) else if (strcmp(kw->name,str)==0) { \
		if (strcmp(kw->value,sy)==0) { conn->policy |= fl; } \
		else if (strcmp(kw->value,sn)==0) { conn->policy &= ~fl; } \
		else { ERR_FOUND("bad policy value %s=%s", kw->name, kw->value) } \
	}

#define KW_POLICY_FLAG_BI(str,sy,sn,fy,fn) else if (strcmp(kw->name,str)==0) { \
		if (strcmp(kw->value,sy)==0) \
			{ conn->policy |= fy; conn->policy &= ~fn; } \
		else if (strcmp(kw->value,sn)==0) \
			{ conn->policy &= ~fy; conn->policy |= fn; } \
		else { ERR_FOUND("bad policy value %s=%s", kw->name, kw->value) } \
	}

#define KW_STR(str,val) else if (strcmp(kw->name,str)==0) { \
		if (val) free(val); \
		val = kw->value ? strdup(kw->value) : NULL; \
	}

#define KW_STR_IN(str,val,in) else if (strcmp(kw->name,str)==0) { \
		if (!kw_in_list(kw->value,in)) \
			ERR_FOUND("bad value %s=%s", kw->name, kw->value) \
		if (val) free(val); \
		val = kw->value ? strdup(kw->value) : NULL; \
	}

#define KW_LIST(str,val) else if (strcmp(kw->name,str)==0) { \
		if (val) _free_list(val); \
		val = _new_list(kw->value); \
	}

#define KW_LIST_IN(str,val,in) else if (strcmp(kw->name,str)==0) { \
		char **_priv_b; \
		if (val) _free_list(val); \
		val = _new_list(kw->value); \
		for (_priv_b=val; (_priv_b && *_priv_b); _priv_b++) \
			if (!kw_in_list(*_priv_b,in)) \
				ERR_FOUND("bad value %s=%s", kw->name, *_priv_b) \
	}

#define KW_INT(str,val) else if (strcmp(kw->name,str)==0) { \
		char *endptr; \
		val = strtoul(kw->value, &endptr, 10); \
		if ((*endptr != '\0') || (endptr == kw->name)) { \
			ERR_FOUND("bad integer value %s=%s", kw->name, kw->value); \
		} \
	}

#define KW_PERCENT(str,val) else if (strcmp(kw->name,str)==0) { \
		char *endptr; \
		val = strtoul(kw->value, &endptr, 10); \
		if (((*endptr != '%') || (endptr[1]!='\0')) || (endptr == kw->name)) { \
			ERR_FOUND("bad percent value %s=%s", kw->name, kw->value); \
		} \
	}

#define KW_DURATION(str,val) else if (strcmp(kw->name,str)==0) { \
		char *endptr; \
		val = strtoul(kw->value, &endptr, 10); \
		if (endptr == kw->name) { \
			ERR_FOUND("bad duration value %s=%s", kw->name, kw->value); \
		} \
		else if (*endptr == '\0') { } \
		else if ((*endptr == 's') && (endptr[1] == '\0')) { } \
		else if ((*endptr == 'm') && (endptr[1] == '\0')) { val *= 60; } \
		else if ((*endptr == 'h') && (endptr[1] == '\0')) { val *= 3600; } \
		else if ((*endptr == 'd') && (endptr[1] == '\0')) { val *= 3600*24; } \
		else { \
			ERR_FOUND("bad duration value %s=%s", kw->name, kw->value); \
		} \
	}

static void _free_list(char **list)
{
	char **s;
	for (s=list; *s; s++) free(*s);
	free(list);
}

static char **_new_list(char *value)
{
	char *val, *b, *e, *end, **ret;
	int count;
	val = value ? strdup(value) : NULL;
	if (!val) return NULL;
	end = val + strlen(val);
	for (b=val, count=0; b<end; ) {
		for (e=b; ((*e!=' ') && (*e!='\0')); e++);
		*e = '\0';
		if (e!=b) { count++; }
		b=e+1;
	}
	if (count==0) {
		free(val);
		return NULL;
	}
	ret = (char **)malloc((count+1) * sizeof(char *));
	if (!ret) {
		free(val);
		return NULL;
	}
	for (b=val, count=0; b<end; ) {
		for (e=b; (*e!='\0'); e++);
		if (e!=b) {
			ret[count++] = strdup(b);
		}
		b=e+1;
	}
	ret[count] = NULL;
	free(val);
	return ret;
}

static int kw_in_list(char *val, char *in_list)
{
	char *in = strdup(in_list), *b, *e, *end;
	if (!in) return 0;
	end = in + strlen(in);
	for (b=in; b<end; ) {
		for (e=b; ((*e!=' ') && (*e!='\0')); e++);
		*e = '\0';
		if (e!=b) {
			if (strcmp(val, b)==0) {
				free(in);
				return 1;
			}
		}
		b=e+1;
	}
	free(in);
	return 0;
}

static int load_setup (struct starter_config *cfg, struct config_parsed *cfgp,
	char **perr)
{
	unsigned int err = 0;
	struct kw_list *kw;

	for (kw=cfgp->config_setup; kw; kw=kw->next) {
		if (0) { }
#ifdef WANT_DEFAULTROUTE
	/* expand KW_LIST macro because %defaultroute changes the value */
		else if (strcmp(kw->name,"interfaces")==0) { 
			if (strcmp(kw->value,"%defaultroute")==0) {
			char _priv[64];
			struct dfltroute _priv_dr;

			if (get_defaultroute(&_priv_dr)) {
				sprintf(_priv, "ipsec0=%s", _priv_dr.phys);
				cfg->setup.interfaces = _new_list(_priv);
				} else {
				ERR_FOUND("No default route - cannot cope.");
				}
			} else {
			if (cfg->setup.interfaces) _free_list(cfg->setup.interfaces);
			cfg->setup.interfaces = _new_list(kw->value);
			}
	}
#else
		KW_LIST("interfaces",cfg->setup.interfaces)
#endif /* WANT_DEFAULTROUTE */
		KW_LIST_IN("klipsdebug",cfg->setup.klipsdebug,
			"tunnel tunnel-xmit pfkey xform eroute spi radij esp ah ipcomp "
			"verbose all none")
		KW_LIST_IN("plutodebug",cfg->setup.plutodebug,
			"raw crypt parsing emitting control klips dns all private none")
		KW_STR("prepluto",cfg->setup.prepluto)
		KW_STR("postpluto",cfg->setup.postpluto)
		KW_BOOL("fragicmp",cfg->setup.fragicmp)
		KW_STR_IN("packetdefault",cfg->setup.packetdefault,"drop reject pass")
		KW_BOOL("hidetos",cfg->setup.hidetos)
		KW_BOOL("uniqueids",cfg->setup.uniqueids)
		KW_INT("overridemtu",cfg->setup.overridemtu)
#ifdef X509_PATCH
		KW_BOOL("strictcrlpolicy",cfg->setup.strictcrlpolicy)
		KW_BOOL("nocrsend",cfg->setup.nocrsend)
#endif
#ifdef NAT_TRAVERSAL
		KW_BOOL("nat_traversal",cfg->setup.nat_traversal)
		KW_INT("keep_alive",cfg->setup.keep_alive)
#endif
#ifdef VIRTUAL_IP
		KW_STR("virtual_private",cfg->setup.virtual_private)
#endif
		else {
			starter_log(LOG_LEVEL_INFO,
				"ignoring unknown keyword '%s' in config setup\n", kw->name);
		}
	}
	return err;
}

static int kw_end (struct starter_conn *conn_st, struct starter_end *end,
	char *origkw, char *kw, char *value, char *conn, char **perr)
{
	unsigned int err = 0;
	err_t er;
#ifdef WANT_DEFAULTROUTE
	struct dfltroute dr;
#endif

	if (strcmp(kw,"")==0) {
#ifdef WANT_DEFAULTROUTE
		if (strcmp(value,"%defaultroute")==0) {
		if ( get_defaultroute (&dr) ) {
			end->iface=strdup(dr.phys);
			if (starter_iface_find(end->iface, AF_INET, &(end->addr),
				&(end->nexthop)) == -1) {
				conn_st->state = STATE_INVALID;
				}
			initaddr ( (char *) &(dr.gw.s_addr), sizeof(dr.gw.s_addr), AF_INET, 
				&(end->nexthop)); 
			} else {
			ERR_FOUND("No default route found - cannot cope.");
			}
		}
#else	
		if (strcmp(value,"%defaultroute")==0) {
			ERR_FOUND("unsupported %s=%s", origkw, value);
		}
#endif
		else if (strcmp(value,"%any")==0) {
			anyaddr(AF_INET, &(end->addr));
		}
		else if (value[0]=='%') {
			if (end->iface) free(end->iface);
			end->iface = strdup(value+1);
			if (starter_iface_find(end->iface, AF_INET, &(end->addr),
				&(end->nexthop)) == -1) {
				conn_st->state = STATE_INVALID;
			}
		}
		else {
			er = ttoaddr(value, 0, AF_INET, &(end->addr));
			if (er) ERR_FOUND("bad addr %s=%s [%s]", origkw, value, er);
		}
	}
	else if (strcmp(kw,"subnet")==0) {
#ifdef VIRTUAL_IP
		if ( ((strlen(value)>=6) && (strncmp(value,"vhost:",6)==0)) ||
			((strlen(value)>=5) && (strncmp(value,"vnet:",5)==0)) ) {
			er = NULL;
			end->virt = strdup(value);
			if (!end->virt) ERR_FOUND("can't alloc memory");
		}
		else {
			end->has_client = TRUE;
			er = ttosubnet(value, 0, AF_INET, &(end->subnet));
		}
#else
		end->has_client = TRUE;
#ifdef X509_PATCH
		end->has_client_wildcard = FALSE;
#endif
		er = ttosubnet(value, 0, AF_INET, &(end->subnet));
#endif
		if (er) ERR_FOUND("bad subnet %s=%s [%s]", origkw, value, er);
	}
	else if (strcmp(kw,"nexthop")==0) {
		er = ttoaddr(value, 0, AF_INET, &(end->nexthop));
		if (er) ERR_FOUND("bad addr %s=%s [%s]", origkw, value, er);
	}
	else if (strcmp(kw,"id")==0) {
		if (end->id) free(end->id);
		end->id = strdup(value);
	}
	else if (strcmp(kw,"rsasigkey")==0) {
		if (end->rsakey) free(end->rsakey);
		end->rsakey = strdup(value);
	}
	else if (strcmp(kw,"updown")==0) {
		if (end->updown) free(end->updown);
		end->updown = strdup(value);
	}
#ifdef X509_PATCH
	else if (strcmp(kw,"cert")==0) {
		if (end->cert) free(end->cert);
		end->cert = strdup(value);
	}
	else if (strcmp(kw,"ca")==0) {
		if (end->ca) free(end->ca);
		end->ca = strdup(value);
	}
	else if (strcmp(kw,"subnetwithin")==0) {
		end->has_client = TRUE;
		end->has_client_wildcard = TRUE;
		er = ttosubnet(value, 0, AF_INET, &(end->subnet));
	}
	else if (strcmp(kw,"protoport")==0) {
	    er = ttoprotoport(value, 0, &(end->protocol), &(end->port), &(end->has_client_wildcard));
	}
#endif
	else {
		starter_log(LOG_LEVEL_INFO,
			"ignoring unknown keyword '%s' in conn %s\n", origkw, conn);
	}
	return err;
}

static int load_conn (struct starter_conn *conn, struct config_parsed *cfgp,
	struct kw_list *kw, char **perr)
{
	unsigned int err = 0;

#ifdef DPD
	char *dpdaction = NULL;
#endif

	for ( ; kw; kw=kw->next) {
		if (strcmp(kw->name,"keyexchange")==0) {
			if (strcmp(kw->value,"ike")!=0)
				ERR_FOUND("bad keyechange value %s=%s", kw->name, kw->value);
		}
		else if (strcmp(kw->name,"auto")==0) {
			if (strcmp(kw->value,"add")==0) conn->startup=STARTUP_ADD;
			else if (strcmp(kw->value,"start")==0) conn->startup=STARTUP_START;
			else if (strcmp(kw->value,"route")==0) conn->startup=STARTUP_ROUTE;
			else if (strcmp(kw->value,"ignore")==0) conn->state=STATE_IGNORE;
			else ERR_FOUND("bad auto value %s=%s", kw->name, kw->value);
		}
		KW_POLICY_FLAG("type","tunnel","transport",POLICY_TUNNEL)
		KW_POLICY_FLAG("pfs","yes","no",POLICY_PFS)
		KW_POLICY_FLAG("compress","yes","no",POLICY_COMPRESS)
		KW_POLICY_FLAG_BI("authby","rsasig","secret",POLICY_RSASIG,POLICY_PSK)
		KW_DURATION("keylife",conn->sa_ipsec_life_seconds)
		KW_DURATION("rekeymargin",conn->sa_rekey_margin)
		KW_DURATION("ikelifetime",conn->sa_ike_life_seconds)
		KW_INT("keyingtries",conn->sa_keying_tries)
		KW_PERCENT("rekeyfuzz",conn->sa_rekey_fuzz)
		KW_POLICY_FLAG("rekey","no","yes",POLICY_DONT_REKEY)
#ifdef ALG_PATCH
		KW_STR("esp",conn->esp)
		KW_STR("ike",conn->ike)
#endif
#ifdef DPD
		KW_INT("dpddelay",conn->dpd_delay)
		KW_INT("dpdtimeout",conn->dpd_timeout)
		KW_STR_IN("dpdaction",dpdaction,"clear hold")
#endif
		else if (strncmp(kw->name,"left",4)==0) {
			err += kw_end(conn, &(conn->left), kw->name, kw->name+4, kw->value,
				conn->name, perr);
		}
		else if (strncmp(kw->name,"right",5)==0) {
			err += kw_end(conn, &(conn->right), kw->name, kw->name+5, kw->value,
				conn->name,perr);
		}
		else {
			starter_log(LOG_LEVEL_INFO,
				"ignoring unknown keyword '%s' in conn %s\n",
				kw->name, conn->name);
		}
	}

#ifdef DPD
	if (dpdaction) {
		if (!strcmp(dpdaction, "clear")) {
        	        conn->dpd_action = DPD_ACTION_CLEAR;
        	} else if (!strcmp(dpdaction, "hold")) {
        	        conn->dpd_action = DPD_ACTION_HOLD;
        	} else {
			/* JJ: Redundant for now... added in case the defaults change somehow */
                	conn->dpd_action = DPD_ACTION_HOLD;
       		}
	}
#endif

	return err;
}

static void conn_default (char *name, struct starter_conn *conn,
	struct starter_conn *def)
{
	struct starter_end *end;
	memcpy(conn, def, sizeof(struct starter_conn));
	conn->name = strdup(name);
#define CONN_STR(v) if (v) v=strdup(v)
	for (end=&(conn->left); ; end = &(conn->right)) {
		CONN_STR(end->iface);
		CONN_STR(end->id);
		CONN_STR(end->rsakey);
		CONN_STR(end->updown);
#ifdef X509_PATCH
		CONN_STR(end->cert);
		CONN_STR(end->ca);
#endif
#ifdef VIRTUAL_IP
		CONN_STR(end->virt);
#endif
		if (end == &(conn->right))
			break;
	}
#ifdef ALG_PATCH
	CONN_STR(conn->esp);
	CONN_STR(conn->ike);
#endif
#undef CONN_STR
}

struct starter_config *confread_load(const char *file, char **perr)
{
	struct starter_config *cfg = NULL;
	struct config_parsed *cfgp;
	struct section_list *sconn;
	struct starter_conn *conn;
	unsigned int err = 0;

	/**
	 * Load file
	 */
	cfgp = parser_load_conf(file, perr);
	if (!cfgp) return NULL;

	cfg = (struct starter_config *)malloc(sizeof(struct starter_config));
	if (!cfg) {
		if (perr) *perr = strdup("can't allocate mem in confread_load()");
		parser_free_conf(cfgp);
		return NULL;
	}

	/**
	 * Set default values
	 */
	default_values(cfg);

	/**
	 * Load setup
	 */
	err += load_setup(cfg, cfgp, perr);

	/**
	 * Find %default conn
	 */
	for (sconn=cfgp->conn_first; (!err) && (sconn); sconn=sconn->next) {
		if (strcmp(sconn->name,"%default")==0) {
			starter_log(LOG_LEVEL_DEBUG, "Loading default conn");
			err += load_conn (&cfg->conn_default, cfgp, sconn->kw, perr);
		}
	}

	/**
	 * Load other conns
	 */
	for (sconn=cfgp->conn_first; (!err) && (sconn); sconn=sconn->next) {
		if (strcmp(sconn->name,"%default")==0) continue;
		starter_log(LOG_LEVEL_DEBUG, "Loading conn %s", sconn->name);
		conn = (struct starter_conn *)malloc(sizeof(struct starter_conn));
		if (!conn) {
			if (perr) *perr = strdup("can't allocate mem in confread_load()");
			parser_free_conf(cfgp);
			confread_free(cfg);
			return NULL;
		}
		conn_default(sconn->name, conn, &cfg->conn_default);
		conn->state = STATE_TO_ADD;
		conn->next = NULL;
		if (cfg->conn_last) cfg->conn_last->next = conn;
		cfg->conn_last = conn;
		if (!cfg->conn_first) cfg->conn_first = conn;
		err += load_conn (conn, cfgp, sconn->kw, perr);
	}

	parser_free_conf(cfgp);

	if (err) {
		confread_free(cfg);
		cfg = NULL;
	}

	return cfg;
}

#define FREE_STR(v) { if (v) { free(v); v=NULL; } }
#define FREE_LST(v) { if (v) { _free_list(v); v=NULL; } }
static void confread_free_conn(struct starter_conn *conn)
{
	struct starter_end *end;
	for (end=&(conn->left); ; end = &(conn->right)) {
		FREE_STR(end->iface);
		FREE_STR(end->id);
		FREE_STR(end->rsakey);
		FREE_STR(end->updown);
#ifdef X509_PATCH
		FREE_STR(end->cert);
		FREE_STR(end->ca);
#endif
#ifdef VIRTUAL_IP
		FREE_STR(end->virt);
#endif
		if (end == &(conn->right))
			break;
	}
#ifdef ALG_PATCH
	FREE_STR(conn->esp);
	FREE_STR(conn->ike);
#endif
	FREE_STR(conn->name);
}

void confread_free(struct starter_config *cfg)
{
	struct starter_conn *conn, *c;
	FREE_LST(cfg->setup.interfaces);
	FREE_LST(cfg->setup.klipsdebug);
	FREE_LST(cfg->setup.plutodebug);
	FREE_STR(cfg->setup.prepluto);
	FREE_STR(cfg->setup.postpluto);
	FREE_STR(cfg->setup.packetdefault);
#ifdef VIRTUAL_IP
	FREE_STR(cfg->setup.virtual_private);
#endif
	confread_free_conn(&(cfg->conn_default));
	for (conn=cfg->conn_first; conn; ) {
		c = conn;
		conn = conn->next;
		confread_free_conn(c);
		free(c);
	}
	free(cfg);
}
#undef FREE_STR

