%{
/* FreeS/WAN config file parser (parser.y)
 * Copyright (C) 2001 Mathieu Lafon - Arkoon Network Security
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * RCSID $Id: parser.y,v 1.2 2002/08/19 16:32:56 dan Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "parser.h"

#define YYERROR_VERBOSE
#define ERRSTRING_LEN	256

/**
 * Bison
 */
static char parser_errstring[ERRSTRING_LEN+1];
void yyerror(const char *s);
extern int yylex (void);

/**
 * Static Globals
 */
static int _save_errors_;
static struct config_parsed *_parser_cfg;
static struct kw_list **_parser_kw, *_parser_kw_last;

/**
 * Functions
 */

%}

%union { char *s; };
%token EQUAL FIRST_SPACES EOL CONFIG SETUP CONN INCLUDE
%token <s> STRING

%%

/*
 * Config file
 */

config_file:
	config_file section_or_include
	| /* NULL */		
	;

section_or_include:
	CONFIG SETUP EOL {
		_parser_kw = &(_parser_cfg->config_setup);
		_parser_kw_last = NULL;
	} kw_section
	| CONN STRING EOL {
		struct section_list *section;
		section = (struct section_list *)malloc(sizeof(struct section_list));
		if (section) {
			section->name = $2;
			section->kw = NULL;
			section->next = NULL;
			_parser_kw = &(section->kw);
			if (!_parser_cfg->conn_first) _parser_cfg->conn_first = section;
			if (_parser_cfg->conn_last) _parser_cfg->conn_last->next = section;
			_parser_cfg->conn_last = section;
			_parser_kw_last = NULL;
		}
		else {
			_parser_kw = NULL;
			_parser_kw_last = NULL;
			yyerror("can't allocate memory in section_or_include/conn");
		}
	} kw_section
	| INCLUDE STRING {
		extern void _parser_y_include (const char *f);
		_parser_y_include($2);
		free($2);
	} EOL
	| EOL
	;

kw_section:
	FIRST_SPACES statement_kw EOL kw_section
	|
	;

statement_kw:
	STRING EQUAL STRING {
		struct kw_list *new;
		if (_parser_kw) {
			new = (struct kw_list *)malloc(sizeof(struct kw_list));
			if (new) {
				new->name = $1;
				new->value = $3;
				new->next = NULL;
				if (_parser_kw_last)
					_parser_kw_last->next = new;
				_parser_kw_last = new;
				if (!*_parser_kw) *_parser_kw = new;
			}
			else {
				yyerror("can't allocate memory in statement_kw");
			}
		}
	}
	| STRING EQUAL { free($1); }
	|
	;

%%

void yyerror(const char *s)
{
	extern void _parser_y_error(char *b, int size, const char *s);
	if (_save_errors_)
		_parser_y_error(parser_errstring, ERRSTRING_LEN, s);
}

struct config_parsed *parser_load_conf (const char *file, char **perr)
{
	struct config_parsed *cfg=NULL;
	int err = 0;
	FILE *f;

	extern void _parser_y_init (const char *f);
	extern FILE *yyin;

	memset(parser_errstring, 0, ERRSTRING_LEN+1);
	if (perr) *perr = NULL;

	cfg = (struct config_parsed *)malloc(sizeof(struct config_parsed));
	if (cfg) {
		memset(cfg, 0, sizeof(struct config_parsed));
		f = fopen(file, "r");
		if (f) {
			yyin = f;
			_parser_y_init(file);
			_save_errors_=1;
			_parser_cfg = cfg;

		    if (yyparse()!=0) {
				if (parser_errstring[0]=='\0') {
					snprintf(parser_errstring, ERRSTRING_LEN,
						"Unknown error...");
				}
				_save_errors_=0;
				while (yyparse()!=0);
				err++;
			}
			else if (parser_errstring[0]!='\0') {
				err++;
			}
			else {
				/**
				 * Config valid
				 */
			}

			fclose(f);
		}
		else {
			snprintf(parser_errstring, ERRSTRING_LEN, "can't load file '%s'",
				file);
			err++;
		}
	}
	else {
		snprintf(parser_errstring, ERRSTRING_LEN, "can't allocate memory");
		err++;
	}

	if (err) {
		if (perr) *perr = strdup(parser_errstring);
		if (cfg) parser_free_conf (cfg);
		cfg = NULL;
	}

	return cfg;
}

static void parser_free_kwlist (struct kw_list *list)
{
	struct kw_list *elt;
	while (list) {
		elt = list;
		list = list->next;
		if (elt->name) free(elt->name);
		if (elt->value) free(elt->value);
		free(elt);
	}
}

void parser_free_conf (struct config_parsed *cfg)
{
	struct section_list *sec;
	if (cfg) {
		parser_free_kwlist(cfg->config_setup);
		while (cfg->conn_first) {
			sec = cfg->conn_first;
			cfg->conn_first = cfg->conn_first->next;
			if (sec->name) free(sec->name);
			parser_free_kwlist(sec->kw);
			free(sec);
		}
		free(cfg);
	}
}

