/* FreeS/WAN whack functions to communicate with pluto (whack.c)
 * Copyright (C) 2001-2002 Mathieu Lafon - Arkoon Network Security
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * RCSID $Id: whack.c,v 1.5 2002/09/03 18:06:03 mathieu Exp $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <linux/stddef.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <openswan.h>

/** FIXME: ugly include lines **/
#include "../pluto/constants.h"
#include "../pluto/defs.h"
#include "../pluto/whack.h"

#ifdef DPD
#define DPD_ACTION_CLEAR	0
#define DPD_ACTION_HOLD		1
#endif

#include "whack.h"
#include "confread.h"
#include "files.h"
#include "log.h"

static int pack_str (char **p, char **next, char **roof)
{
	const char *s = (*p==NULL) ? "" : *p;    /* note: NULL becomes ""! */
	size_t len = strlen(s)+1;

	if ((*roof - *next) < len) {
		/** not enough space **/
		return 0;
	}
	else {
		strcpy(*next, s);
		(*next) += len;
		*p = NULL;   /* don't send pointers on the wire! */
		return 1;
	}
}

static int send_whack_msg (struct whack_message *msg)
{
	struct sockaddr_un ctl_addr = { AF_UNIX, CTL_FILE };
	int sock;
	ssize_t len;
	char *str_next, *str_roof;

	/**
	 * Pack strings
	 */
	str_next = (char *)msg->string;
	str_roof = (char *)&msg->string[sizeof(msg->string)];
	if (!pack_str(&msg->name, &str_next, &str_roof)                 /* string 1 */
		|| !pack_str(&msg->left.id, &str_next, &str_roof)       /* string 2 */
#ifdef X509_PATCH
		|| !pack_str(&msg->left.cert, &str_next, &str_roof)     /* string 3 */
		|| !pack_str(&msg->left.ca, &str_next, &str_roof)       /* string 4 */
#endif
		|| !pack_str(&msg->left.updown, &str_next, &str_roof)   /* string 5 */
#ifdef VIRTUAL_IP
		|| !pack_str(&msg->left.virt, &str_next, &str_roof)
#endif
		|| !pack_str(&msg->right.id, &str_next, &str_roof)      /* string 6 */
#ifdef X509_PATCH
		|| !pack_str(&msg->right.cert, &str_next, &str_roof)    /* string 7 */
		|| !pack_str(&msg->right.ca, &str_next, &str_roof)      /* string 8 */
#endif
		|| !pack_str(&msg->right.updown, &str_next, &str_roof)  /* string 9 */
#ifdef VIRTUAL_IP
		|| !pack_str(&msg->right.virt, &str_next, &str_roof)
#endif
		|| !pack_str(&msg->keyid, &str_next, &str_roof)         /* string 10 */
#ifdef ALG_PATCH
		|| !pack_str(&msg->ike, &str_next, &str_roof)           /* string 11 */
		|| !pack_str(&msg->esp, &str_next, &str_roof)           /* string 12 */
#endif
		|| (str_roof - str_next < msg->keyval.len)) {
		starter_log(LOG_LEVEL_ERR, "send_wack_msg(): can't pack strings");
		return -1;
	}
	if (msg->keyval.ptr) memcpy(str_next, msg->keyval.ptr, msg->keyval.len);
	msg->keyval.ptr = NULL;
	str_next += msg->keyval.len;
	len = str_next - (char *)msg;

	/**
	 * Connect to pluto ctl
	 */
	sock = socket(AF_UNIX, SOCK_STREAM, 0);
	if (sock < 0) {
		starter_log(LOG_LEVEL_ERR, "socket() failed: %s", strerror(errno));
		return -1;
	}
	if (connect(sock, (struct sockaddr *)&ctl_addr,
		offsetof(struct sockaddr_un, sun_path) + strlen(ctl_addr.sun_path))<0) {
		starter_log(LOG_LEVEL_ERR, "connect(pluto_ctl) failed: %s",
			strerror(errno));
		close(sock);
		return -1;
	}

	/**
	 * Send message
	 */
	if (write(sock, msg, len) != len) {
		starter_log(LOG_LEVEL_ERR, "write(pluto_ctl) failed: %s",
			strerror(errno));
		close(sock);
		return -1;
	}

	/**
	 * TODO: read reply
	 */

	close(sock);
	return 0;
}

static void init_whack_msg (struct whack_message *msg)
{
	memset(msg, 0, sizeof(struct whack_message));
	msg->magic = WHACK_MAGIC;
}

static char *connection_name (struct starter_conn *conn)
{
	/**
	 * If connection name is '%auto', create a new name like conn_xxxxx
	 */
	static char buf[32];
	if (strcmp(conn->name, "%auto")==0) {
		sprintf(buf, "conn_%ld", conn->id);
		return buf;
	}
	else {
		return conn->name;
	}
	return conn->name;
}

static void set_whack_end(struct whack_end *w, struct starter_end *l)
{
	w->id = l->id;
	w->host_addr = l->addr;
	w->host_nexthop = l->nexthop;
	w->has_client = l->has_client;
	if (l->has_client) {
		w->client = l->subnet;
	}
	else {
		w->client.addr.u.v4.sin_family = AF_INET;
	}
	w->updown = l->updown;
	w->host_port = IKE_UDP_PORT;
#ifdef X509_PATCH
	w->has_client_wildcard = l->has_client_wildcard;
	w->cert = l->cert;
	w->protocol = l->protocol;
	w->port = l->port;
#endif
#ifdef VIRTUAL_IP
	w->virt = l->virt;
#endif
}

static int starter_whack_add_pubkey (struct starter_conn *conn,
	struct starter_end *end, const char *lr)
{
	const char *err;
	static char keyspace[1024 + 4];
	struct whack_message msg;

	init_whack_msg(&msg);

	msg.whack_key = TRUE;
	msg.pubkey_alg = PUBKEY_ALG_RSA;
	if (end->id && end->rsakey) {
		/* special values to ignore */
		if ((strcmp(end->rsakey, "")==0) ||
			(strcmp(end->rsakey, "%none")==0) ||
			(strcmp(end->rsakey, "%cert")==0) ||
			(strcmp(end->rsakey, "0x00")==0)) {
			return 0;
		}
		msg.keyid = end->id;
		err = atobytes(end->rsakey, 0, keyspace, sizeof(keyspace),
			&msg.keyval.len);
		if (err) {
			starter_log(LOG_LEVEL_ERR, "conn %s/%s: rsakey malformed [%s]",
				connection_name(conn), lr, err);
			return 1;
		}
		else {
			msg.keyval.ptr = keyspace;
			return send_whack_msg(&msg);
		}
	}
	return 0;
}

int starter_whack_add_conn (struct starter_conn *conn)
{
	struct whack_message msg;
	int r;

	init_whack_msg(&msg);

	msg.whack_connection = TRUE;
	msg.name = connection_name(conn);

	msg.addr_family = AF_INET;
	msg.tunnel_addr_family = AF_INET;

	msg.sa_ike_life_seconds = conn->sa_ike_life_seconds;
	msg.sa_ipsec_life_seconds = conn->sa_ipsec_life_seconds;
	msg.sa_rekey_margin = conn->sa_rekey_margin;
	msg.sa_rekey_fuzz = conn->sa_rekey_fuzz;
	msg.sa_keying_tries = conn->sa_keying_tries;

	msg.policy = conn->policy;

	set_whack_end(&msg.left, &conn->left);
	set_whack_end(&msg.right, &conn->right);

#ifdef ALG_PATCH
	msg.esp = conn->esp;
	msg.ike = conn->ike;
#endif
#ifdef DPD
	msg.dpd_timeout = conn->dpd_timeout;
	msg.dpd_delay = conn->dpd_delay;
	msg.dpd_action = conn->dpd_action;
#endif

	r =  send_whack_msg(&msg);

	if ((r==0) && (conn->policy & POLICY_RSASIG)) {
		r += starter_whack_add_pubkey (conn, &conn->left, "left");
		r += starter_whack_add_pubkey (conn, &conn->right, "right");
	}

	return r;
}

int starter_whack_del_conn (struct starter_conn *conn)
{
	struct whack_message msg;
	init_whack_msg(&msg);
	msg.whack_delete = TRUE;
	msg.name = connection_name(conn);
	return send_whack_msg(&msg);
}

int starter_whack_route_conn (struct starter_conn *conn)
{
	struct whack_message msg;
	init_whack_msg(&msg);
	msg.whack_route = TRUE;
	msg.name = connection_name(conn);
	return send_whack_msg(&msg);
}

int starter_whack_initiate_conn (struct starter_conn *conn)
{
	struct whack_message msg;
	init_whack_msg(&msg);
	msg.whack_initiate = TRUE;
	msg.whack_async = TRUE;
	msg.name = connection_name(conn);
	return send_whack_msg(&msg);
}

int starter_whack_listen (void)
{
	struct whack_message msg;
	init_whack_msg(&msg);
	msg.whack_listen = TRUE;
	return send_whack_msg(&msg);
}

int starter_whack_shutdown (void)
{
	struct whack_message msg;
	init_whack_msg(&msg);
	msg.whack_shutdown = TRUE;
	return send_whack_msg(&msg);
}

